package ru.dbotthepony.mc.otm.client.screen

import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Items
import ru.dbotthepony.mc.otm.block.entity.storage.ItemMonitorPlayerSettings
import ru.dbotthepony.mc.otm.client.render.SkinGrid
import ru.dbotthepony.mc.otm.client.render.UVWindingOrder
import ru.dbotthepony.mc.otm.client.screen.panels.*
import ru.dbotthepony.mc.otm.client.screen.widget.ProgressGaugePanel
import ru.dbotthepony.mc.otm.menu.ItemMonitorMenu
import ru.dbotthepony.mc.otm.next
import ru.dbotthepony.mc.otm.prev

private fun div(a: Int, b: Int): Int {
	if (a % b == 0) {
		return a / b - 1
	}

	return a / b
}

class ItemMonitorScreen(menu: ItemMonitorMenu, inventory: Inventory, title: Component) :
	MatteryScreen<ItemMonitorMenu>(menu, inventory, title) {

	override fun makeMainFrame(): FramePanel {
		val frame = FramePanel(this@ItemMonitorScreen, null, 0f, 0f, FRAME_WIDTH, FRAME_HEIGHT, getTitle())

		val topPanel = EditablePanel(this, frame)
		topPanel.height = ITEM_GRID_HEIGHT * 18f
		topPanel.dock = Dock.TOP

		val bottomPanel = EditablePanel(this, frame)
		bottomPanel.height = 3 * 18f
		bottomPanel.dock = Dock.TOP
		bottomPanel.setDockMargin(top = 6f)

		frame.height = topPanel.height + bottomPanel.height + frame.dockPadding.top + frame.dockPadding.bottom + 6f
		frame.width = 178f + frame.dockPadding.left + frame.dockPadding.right

		val viewScrollBar = ScrollBarPanel(this, topPanel, 28f + ITEM_GRID_WIDTH * 18f + 2f, 16f, ITEM_GRID_HEIGHT * 18f)
		viewScrollBar.setupRowMultiplier { div(menu.view.itemCount, ITEM_GRID_WIDTH) }

		viewScrollBar.dock = Dock.RIGHT
		viewScrollBar.setDockMargin(left = 2f)

		val gridPanel = GridPanel(this, topPanel, width = ITEM_GRID_WIDTH * 18f, height = ITEM_GRID_HEIGHT * 18f, columns = ITEM_GRID_WIDTH, rows = ITEM_GRID_HEIGHT)
		gridPanel.dock = Dock.FILL

		for (i in 0 until ITEM_GRID_WIDTH * ITEM_GRID_HEIGHT) {
			object : AbstractSlotPanel(this@ItemMonitorScreen, gridPanel) {
				override fun getItemStack(): ItemStack {
					val index = i + viewScrollBar.getScroll(div(menu.view.itemCount, ITEM_GRID_WIDTH)) * ITEM_GRID_WIDTH
					return menu.view.sortedView.getOrNull(index)?.stack ?: ItemStack.EMPTY
				}

				override fun mouseScrolledInner(mouse_x: Double, mouse_y: Double, scroll: Double): Boolean {
					return viewScrollBar.mouseScrolledInner(mouse_x, mouse_y, scroll)
				}

				override fun mouseClickedInner(mouse_x: Double, mouse_y: Double, mouse_click_type: Int): Boolean {
					val index = i + viewScrollBar.getScroll(div(menu.view.itemCount, ITEM_GRID_WIDTH)) * ITEM_GRID_WIDTH
					menu.view.mouseClick(index, mouse_click_type)
					return true
				}
			}
		}

		val craftingGrid = GridPanel(this, bottomPanel, width = 3 * 18f, height = 3 * 18f, columns = 3, rows = 3)
		craftingGrid.dock = Dock.LEFT

		for (i in 0 until 9) {
			SlotPanel(this, craftingGrid, menu.craftingSlots[i])
		}

		val arrowAndButtons = object : EditablePanel(this@ItemMonitorScreen, bottomPanel, width = ProgressGaugePanel.GAUGE_BACKGROUND.w) {
			override fun innerRender(stack: PoseStack, mouse_x: Float, mouse_y: Float, flag: Float) {
				ProgressGaugePanel.GAUGE_BACKGROUND.render(stack, y = height / 2f - ProgressGaugePanel.GAUGE_BACKGROUND.h / 2f)
			}
		}

		arrowAndButtons.dock = Dock.LEFT
		arrowAndButtons.setDockMargin(left = 4f)

		val arrowLine = EditablePanel(this, arrowAndButtons, y = 8f, height = 8f, width = arrowAndButtons.width)

		val refillPriority = SmallEnumSquareButtonPanel(this, arrowLine,
			enum = ItemMonitorPlayerSettings.RefillSource::class.java,
			prop = menu.settings::refillSource,
			defaultValue = ItemMonitorPlayerSettings.RefillSource.SYSTEM,
			lambdaOnChange = { menu.sendSettingsToServer() })

		refillPriority.mainTooltip = TranslatableComponent("otm.gui.item_monitor.refill_source.desc")
		refillPriority.add(ItemMonitorPlayerSettings.RefillSource.SYSTEM, ItemMonitorPlayerSettings.RefillSource.SYSTEM.component, ARROW, UVWindingOrder.FLIP)
		refillPriority.add(ItemMonitorPlayerSettings.RefillSource.INVENTORY, ItemMonitorPlayerSettings.RefillSource.INVENTORY.component, ARROW)
		refillPriority.add(ItemMonitorPlayerSettings.RefillSource.INVENTORY_FIRST, ItemMonitorPlayerSettings.RefillSource.INVENTORY_FIRST.component, ARROW_UD, UVWindingOrder.FLIP)
		refillPriority.add(ItemMonitorPlayerSettings.RefillSource.SYSTEM_FIRST, ItemMonitorPlayerSettings.RefillSource.SYSTEM_FIRST.component, ARROW_UD)
		refillPriority.add(ItemMonitorPlayerSettings.RefillSource.DO_NOT, ItemMonitorPlayerSettings.RefillSource.DO_NOT.component, MINUS)

		refillPriority.setDockMargin(2f)

		refillPriority.dock = Dock.LEFT

		val resultAndButtons = EditablePanel(this, bottomPanel, width = 18f)

		resultAndButtons.dock = Dock.LEFT
		resultAndButtons.setDockMargin(left = 6f)

		SlotPanel(this, resultAndButtons, menu.craftingResult, y = 18f)

		val resultTarget = SmallEnumSquareButtonPanel(this, resultAndButtons, y = 8f,
			enum = ItemMonitorPlayerSettings.ResultTarget::class.java,
			prop = menu.settings::resultTarget,
			defaultValue = ItemMonitorPlayerSettings.ResultTarget.MIXED,
			lambdaOnChange = { menu.sendSettingsToServer() })

		resultTarget.mainTooltip = TranslatableComponent("otm.gui.item_monitor.result_target.desc")
		resultTarget.add(ItemMonitorPlayerSettings.ResultTarget.MIXED, ItemMonitorPlayerSettings.ResultTarget.MIXED.component, ARROW_UD)
		resultTarget.add(ItemMonitorPlayerSettings.ResultTarget.ALL_INVENTORY, ItemMonitorPlayerSettings.ResultTarget.ALL_INVENTORY.component, ARROW_UU, UVWindingOrder.FLIP)
		resultTarget.add(ItemMonitorPlayerSettings.ResultTarget.ALL_SYSTEM, ItemMonitorPlayerSettings.ResultTarget.ALL_SYSTEM.component, ARROW_UU)

		val craftingAmount = SmallEnumSquareButtonPanel(this, resultAndButtons, x = 10f, y = 8f,
			enum = ItemMonitorPlayerSettings.Amount::class.java,
			prop = menu.settings::craftingAmount,
			defaultValue = ItemMonitorPlayerSettings.Amount.STACK,
			lambdaOnChange = { menu.sendSettingsToServer() })

		craftingAmount.mainTooltip = TranslatableComponent("otm.gui.item_monitor.amount.desc")
		craftingAmount.add(ItemMonitorPlayerSettings.Amount.ONE, ItemMonitorPlayerSettings.Amount.ONE.component, ONE)
		craftingAmount.add(ItemMonitorPlayerSettings.Amount.STACK, ItemMonitorPlayerSettings.Amount.STACK.component, S)
		craftingAmount.add(ItemMonitorPlayerSettings.Amount.FULL, ItemMonitorPlayerSettings.Amount.FULL.component, F)

		val craftingHistory = GridPanel(this, bottomPanel, width = 3 * 18f, height = 3 * 18f, columns = 3, rows = 3)
		craftingHistory.dock = Dock.LEFT
		craftingHistory.setDockMargin(left = 4f)

		val craftingHistoryScroll = ScrollBarPanel(this, bottomPanel, 0f, 0f, 0f)
		craftingHistoryScroll.dock = Dock.LEFT
		craftingHistoryScroll.setDockMargin(left = 2f)

		for (i in 0 until 9) {
			object : AbstractSlotPanel(this@ItemMonitorScreen, craftingHistory) {
				override fun getItemStack(): ItemStack {
					return ItemStack(Items.ARROW, 42)
				}

				override fun mouseScrolledInner(mouse_x: Double, mouse_y: Double, scroll: Double): Boolean {
					return craftingHistoryScroll.mouseScrolledInner(mouse_x, mouse_y, scroll)
				}
			}
		}

		return frame
	}

	companion object {
		const val FRAME_WIDTH = 240f
		const val FRAME_HEIGHT = 210f

		const val ITEM_GRID_WIDTH = 9
		const val ITEM_GRID_HEIGHT = 5

		val ARROW = SkinGrid.WIDGETS_8[1, 0]
		val ARROW_UD = SkinGrid.WIDGETS_8[1, 1]
		val ARROW_UU = SkinGrid.WIDGETS_8[1, 2]
		val MINUS = SkinGrid.WIDGETS_8[1, 3]

		val S = SkinGrid.WIDGETS_8[2, 0]
		val F = SkinGrid.WIDGETS_8[3, 0]
		val E = SkinGrid.WIDGETS_8[4, 0]
		val ONE = SkinGrid.WIDGETS_8[5, 0]
	}
}
