package ru.dbotthepony.mc.otm.menu

import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.*
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraftforge.network.PacketDistributor
import ru.dbotthepony.mc.otm.container.ItemFilter
import ru.dbotthepony.mc.otm.container.ItemFilterNetworkSlot
import ru.dbotthepony.mc.otm.menu.data.MultiByteDataContainer
import ru.dbotthepony.mc.otm.menu.widget.AbstractWidget
import ru.dbotthepony.mc.otm.network.MatteryNetworking
import java.util.Collections

abstract class MatteryMenu @JvmOverloads protected constructor(
	p_38851_: MenuType<*>?,
	p_38852_: Int,
	val inventory: Inventory,
	val tile: BlockEntity? = null
) : AbstractContainerMenu(p_38851_, p_38852_) {
	val ply: Player get() = inventory.player

	@JvmField
	val matteryWidgets = ArrayList<AbstractWidget>()

	@JvmField
	val inventorySlots = ArrayList<MatterySlot>()
	val multiByteContainers = ArrayList<MultiByteDataContainer>()

	@JvmField
	protected val lockedInventorySlots: MutableSet<Int> = HashSet()
	protected open fun isInventorySlotLocked(index: Int): Boolean = lockedInventorySlots.contains(index)

	@JvmField
	protected var _synchronizer: ContainerSynchronizer? = null

	private val _filterSlots = ArrayList<ItemFilterNetworkSlot>()
	val filterSlots = Collections.unmodifiableList(_filterSlots)

	fun addFilterSlots(slots: ItemFilter): List<ItemFilterNetworkSlot> {
		val result = ArrayList<ItemFilterNetworkSlot>(slots.size)

		for (i in 0 until slots.size) {
			_filterSlots.add(ItemFilterNetworkSlot(i, _filterSlots.size, slots).also(result::add))
		}

		return result
	}

	fun addFilterSlots(amount: Int): List<ItemFilterNetworkSlot> {
		val result = ArrayList<ItemFilterNetworkSlot>(amount)

		for (i in 0 until amount) {
			_filterSlots.add(ItemFilterNetworkSlot(i, _filterSlots.size, null).also(result::add))
		}

		return result
	}

	override fun setSynchronizer(p_150417_: ContainerSynchronizer) {
		_synchronizer = p_150417_
		super.setSynchronizer(p_150417_)
	}

	fun addWidget(widget: AbstractWidget): Int {
		val indexOf = matteryWidgets.indexOf(widget)

		if (indexOf == -1) {
			matteryWidgets.add(widget)
			return matteryWidgets.size - 1
		}

		return indexOf
	}

	fun getWidget(index: Int): AbstractWidget? {
		if (index !in 0 until matteryWidgets.size) {
			return null
		}

		return matteryWidgets[index]
	}

	fun addDataContainer(container: MultiByteDataContainer) {
		if (multiByteContainers.contains(container))
			throw IllegalStateException("Already containing $container")

		multiByteContainers.add(container)
		container.slotID = multiByteContainers.size - 1
	}

	fun getDataContainer(index: Int): MultiByteDataContainer? {
		if (index !in 0 until multiByteContainers.size) {
			return null
		}

		return multiByteContainers[index]
	}

	@JvmField
	protected var inventorySlotIndexStart = 0

	@JvmField
	protected var inventorySlotIndexEnd = 0

	protected fun addInventorySlots() {
		var first = true

		for (i in 0..2) {
			for (j in 0..8) {
				val slot: MatterySlot = object : MatterySlot(inventory, j + i * 9 + 9, 8 + j * 18, 14 + i * 18) {
					override fun mayPlace(p_40231_: ItemStack): Boolean {
						return !isInventorySlotLocked(index)
					}

					override fun mayPickup(p_40228_: Player): Boolean {
						return !isInventorySlotLocked(index)
					}
				}

				inventorySlots.add(slot)
				addSlot(slot)

				if (first) {
					first = false
					inventorySlotIndexStart = slot.index
				}
			}
		}

		var last: MatterySlot? = null

		for (k in 0..8) {
			last = object : MatterySlot(inventory, k, 8 + k * 18, 14 + 58) {
				override fun mayPlace(p_40231_: ItemStack): Boolean {
					return !isInventorySlotLocked(index)
				}

				override fun mayPickup(p_40228_: Player): Boolean {
					return !isInventorySlotLocked(index)
				}
			}

			addSlot(last)
			inventorySlots.add(last)
		}

		inventorySlotIndexEnd = last!!.index
	}

	private val pdistributor = PacketDistributor.PLAYER.with { ply as ServerPlayer }

	override fun broadcastChanges() {
		for (widget in matteryWidgets) {
			widget.updateServer()
		}

		for (data in multiByteContainers) {
			if (data.dirty) {
				data.dirty = false
				MatteryNetworking.CHANNEL.send(pdistributor, data.makePacket())
			}
		}

		super.broadcastChanges()

		val consumer = PacketDistributor.PLAYER.with { ply as ServerPlayer }

		for (slot in _filterSlots) {
			val packet = slot.sendChanges()

			if (packet != null) {
				MatteryNetworking.CHANNEL.send(consumer, packet)
			}
		}
	}

	override fun broadcastFullState() {
		for (widget in matteryWidgets) {
			widget.updateServer()
		}

		for (data in multiByteContainers) {
			MatteryNetworking.CHANNEL.send(pdistributor, data.makePacket())
		}

		super.broadcastFullState()

		val consumer = PacketDistributor.PLAYER.with { ply as ServerPlayer }

		for (slot in _filterSlots) {
			val packet = slot.sendChanges(true)

			if (packet != null) {
				MatteryNetworking.CHANNEL.send(consumer, packet)
			}
		}
	}

	override fun stillValid(player: Player): Boolean {
		if (tile == null) return true

		if (player.level.getBlockEntity(tile.blockPos) !== tile) {
			return false
		}

		val pos = tile.blockPos

		return player.distanceToSqr(pos.x.toDouble() + 0.5, pos.y.toDouble() + 0.5, pos.z.toDouble() + 0.5) <= 64.0
	}

	protected abstract fun getWorkingSlotStart(): Int
	protected abstract fun getWorkingSlotEnd(): Int

	// This method receive Player interactor and slot_index where Shift + Right click occurred
	// It shall return item stack that got moved
	override fun quickMoveStack(ply: Player, slot_index: Int): ItemStack {
		// this.moveItemStackTo(ItemStack, int start_slot_index, int end_slot_index, boolean iteration_order)
		// returns boolean, telling whenever ItemStack was modified (moved or shrank)
		// false means nothing happened
		// Last boolean determine order of slot iteration, where:
		// if TRUE, iteration order is end_slot_index -> start_slot_index
		// if FALSE iteration order is start_slot_index -> end_slot_index
		val start = getWorkingSlotStart()
		val end = getWorkingSlotEnd()

		if (start == end) {
			return ItemStack.EMPTY
		}

		var moved = ItemStack.EMPTY
		val initialSlot = slots[slot_index]

		if (initialSlot.hasItem()) {
			val initialItem = initialSlot.item
			moved = initialItem.copy()

			if (slot_index < inventorySlotIndexStart) {
				// Moving FROM machine TO inventory
				if (!moveItemStackTo(initialItem, inventorySlotIndexStart, inventorySlotIndexEnd + 1, false)) {
					return ItemStack.EMPTY
				}
			} else if (!moveItemStackTo(initialItem, start, end, false)) {
				// Moving FROM inventory TO machine
				return ItemStack.EMPTY
			}

			if (initialItem.isEmpty) {
				initialSlot.set(ItemStack.EMPTY)
			} else {
				initialSlot.setChanged()
			}
		}

		return moved
	}

	fun quickMoveToInventory(stack: ItemStack, simulate: Boolean): ItemStack {
		if (inventorySlotIndexStart == 0 && inventorySlotIndexEnd == 0) {
			return stack
		}

		return moveItemStackToSlots(stack, inventorySlotIndexStart, inventorySlotIndexEnd, simulate = simulate)
	}

	override fun moveItemStackTo(
		item: ItemStack,
		initialSlot: Int,
		finalSlot: Int,
		inverse: Boolean
	): Boolean {
		val remainder = moveItemStackToSlots(item, initialSlot, finalSlot, inverse)

		if (remainder.count == item.count) {
			return false
		}

		item.count = remainder.count
		return true
	}

	fun moveItemStackToSlots(item: ItemStack, initialSlot: Int, finalSlot: Int, inverse: Boolean = false, simulate: Boolean = false): ItemStack {
		if (initialSlot > finalSlot) {
			return item
		}

		val copy = item.copy()

		// first pass - stack with existing slots
		if (copy.isStackable) {
			for (i in (if (inverse) finalSlot downTo initialSlot else initialSlot .. finalSlot)) {
				val slot = slots[i]
				val limit = slot.getMaxStackSize(copy)

				if (limit > slot.item.count && ItemStack.isSameItemSameTags(slot.item, copy)) {
					val newCount = (slot.item.count + copy.count).coerceAtMost(limit)
					val diff = newCount - slot.item.count
					copy.count -= diff

					if (!simulate) {
						slot.item.count += diff
						slot.setChanged()
					}

					if (copy.isEmpty) {
						return copy
					}
				}
			}
		}

		// second pass - drop stack into first free slot
		for (i in (if (inverse) finalSlot downTo initialSlot else initialSlot .. finalSlot)) {
			val slot = slots[i]
			val limit = slot.getMaxStackSize(copy)

			if (!slot.hasItem()) {
				val newCount = copy.count.coerceAtMost(limit)

				if (!simulate) {
					slot.set(copy.copy().also { it.count = newCount })
					slot.setChanged()
				}

				copy.count -= newCount

				if (copy.isEmpty) {
					return copy
				}
			}
		}

		return copy
	}

	public override fun addDataSlots(p_38885_: ContainerData) {
		super.addDataSlots(p_38885_)
	}
}
